/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.Month;
import java.time.temporal.TemporalAmount;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;

class AgeCalculatorTest {
    @ParameterizedTest
    @MethodSource("isUnder18Arguments")
    void testIsUnder18(LocalDate birthDate, LocalDateTime now, boolean expectedUnder18) {
        assertEquals(expectedUnder18, AgeCalculator.isUnder18(birthDate, now));
    }

    private static Stream<Arguments> isUnder18Arguments() {
        final LocalDateTime now = LocalDateTime.of(2021, Month.DECEMBER, 9, 16, 14);
        return Stream.of(
            Arguments.of(LocalDate.of(2001, 1, 1), now, false), // over 18
            Arguments.of(LocalDate.of(2003, 12, 9), now, false), // 18th birthday today
            Arguments.of(LocalDate.of(2003, 12, 10), now, true),  // 18th birthday tomorrow
            Arguments.of(LocalDate.of(1903, 12, 10), now, false)  // born 1903 (118th birthday tomorrow)
        );
    }

    @ParameterizedTest
    @MethodSource("timeUntil18Arguments")
    void testTimeUntil18(LocalDate birthDate, LocalDateTime now, TemporalAmount expectedTimeUntil18) {

        final TemporalAmount timeUntil18 = AgeCalculator.timeUntil18(birthDate, now);

        assertEquals(expectedTimeUntil18, timeUntil18);
    }

    private static Stream<Arguments> timeUntil18Arguments() {
        final LocalDateTime now = LocalDateTime.of(2021, Month.DECEMBER, 9, 16, 14); // 7 hours and 46 minutes before midnight
        return Stream.of(
            Arguments.of(LocalDate.of(2001, 1, 1), now, Duration.ZERO), //over 18
            Arguments.of(LocalDate.of(2003, 12, 9), now, Duration.ZERO), // 18th birthday today
            Arguments.of(LocalDate.of(1903, 12, 10), now, Duration.ZERO),  // born 1903 (118th birthday tomorrow)

            // the following durations are calculated as: duration = 24H * [days until 18th birthday] + [time until midnight]
            Arguments.of(LocalDate.of(2003, 12, 10), now, Duration.ofHours(7).plusMinutes(46)),    // 18th birthday tomorrow   - 24H *  0 + 7H + 46M =   7H + 46M
            Arguments.of(LocalDate.of(2004, 1, 4), now, Duration.ofHours(607).plusMinutes(46)),  // 18th birthday in 25 days - 24H * 25 + 7H + 46M = 607H + 46M
            Arguments.of(LocalDate.of(2004, 1, 5), now, Duration.ofHours(631).plusMinutes(46)),  // 18th birthday in 26 days - 24H * 26 + 7H + 46M = 631H + 46M
            Arguments.of(LocalDate.of(2004, 1, 6), now, Duration.ofHours(655).plusMinutes(46))   // 18th birthday in 27 days - 24H * 27 + 7H + 46M = 655H + 46M
        );
    }

}
